///////////////////////////////////////////////////////////////////////
// DiscFormatData.cpp
//
// Wrapper for IDiscFormat2Data interface

#include "stdafx.h"
#include "res/resource.h"
#include "DiscFormatData.h"


CDiscFormatData::CDiscFormatData(void) :
    pDF2Data(NULL),
    lastError(S_OK),
    hNotificationWnd(NULL),
    pSupportedMediaTypes(NULL),
    pSupportedWriteSpeeds(NULL)
{
}

CDiscFormatData::~CDiscFormatData(void)
{
	if(pSupportedMediaTypes)
		SafeArrayDestroy(pSupportedMediaTypes);
	if(pSupportedWriteSpeeds)
		SafeArrayDestroy(pSupportedWriteSpeeds);
	if(pDF2Data)
		pDF2Data->Release();
}

bool CDiscFormatData::Initialize(CDiscRecorder* pDiscRecorder, const CString& clientName, HWND hNotificationWnd)
{
	ASSERT(pDF2Data == NULL);
	this->hNotificationWnd = hNotificationWnd;

	CString lastErrorMessage;
	HRESULT lastError = CoCreateInstance(__uuidof(MsftDiscFormat2Data), NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&pDF2Data));
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_INIT_DISCFORMAT2DATA, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}

	VARIANT_BOOL isSupported = VARIANT_FALSE;
	lastError = pDF2Data->IsRecorderSupported(pDiscRecorder->GetInterface(), &isSupported);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_ISRECORDERSUPPORTED_DATA, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	if(isSupported == VARIANT_FALSE) {
		lastErrorMessage.LoadString(IDS_ERR_RECORDER_NOT_SUPPORTED);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	lastError = pDF2Data->put_Recorder(pDiscRecorder->GetInterface());
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_PUTRECORDER_DATA, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	CComBSTR bstr = clientName;
	lastError = pDF2Data->put_ClientName(bstr);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_PUTCLIENTNAME_DATA, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}

	SetLastError(lastError);
	return true;
}

void CDiscFormatData::SetLastError(HRESULT lastError)
{
	this->lastError = lastError;
}

void CDiscFormatData::SetLastError(HRESULT lastError, CString lastErrorMessage)
{
	this->lastError = lastError;
	this->lastErrorMessage = lastErrorMessage;
}

IMAPI_MEDIA_PHYSICAL_TYPE CDiscFormatData::GetCurrentPhysicalMediaType()
{
	ASSERT(pDF2Data != NULL);
	CString lastErrorMessage;

	IMAPI_MEDIA_PHYSICAL_TYPE t = IMAPI_MEDIA_TYPE_UNKNOWN;
	HRESULT lastError = pDF2Data->get_CurrentPhysicalMediaType(&t);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETPHYSICALMEDIATYPE, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);

	return t;
}

bool CDiscFormatData::IsCurrentMediaSupported(CDiscRecorder* pDiscRecorder)
{
	ASSERT(pDF2Data != NULL);
	CString lastErrorMessage;

	VARIANT_BOOL b = VARIANT_FALSE;
	HRESULT lastError = pDF2Data->IsCurrentMediaSupported(pDiscRecorder->GetInterface(), &b);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_ISMEDIASUPPORTED_DATA, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);

	return (b != VARIANT_FALSE);
}

BOOL CDiscFormatData::IsMediaBlank()
{
	ASSERT(pDF2Data != NULL);

	VARIANT_BOOL b = VARIANT_FALSE;
	HRESULT lastError = pDF2Data->get_MediaHeuristicallyBlank(&b);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETMEDIAHEURISTICALLYBLANK_DATA, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);

	return (b != VARIANT_FALSE);
}

ULONG CDiscFormatData::GetTotalSupportedMediaTypes()
{
	ASSERT(pDF2Data != NULL);
	CString lastErrorMessage;

	if(pSupportedMediaTypes) {
		SafeArrayDestroy(pSupportedMediaTypes);
		pSupportedMediaTypes = NULL;
	}

	HRESULT lastError = pDF2Data->get_SupportedMediaTypes(&pSupportedMediaTypes);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETSUPPORTEDMEDIATYPES_DATA, lastError);
		SetLastError(lastError, lastErrorMessage);
		return 0;
	}
	SetLastError(lastError);
	return pSupportedMediaTypes->rgsabound[0].cElements;
}

IMAPI_MEDIA_PHYSICAL_TYPE CDiscFormatData::GetSupportedMediaType(LONG index)
{
	if(!pSupportedMediaTypes)
		GetTotalSupportedMediaTypes();
	ASSERT(pSupportedMediaTypes != NULL);

	IMAPI_MEDIA_PHYSICAL_TYPE mediaType = IMAPI_MEDIA_TYPE_UNKNOWN;
	if(pSupportedMediaTypes) {
		VARIANT v;
		SafeArrayGetElement(pSupportedMediaTypes, &index, reinterpret_cast<LPVOID>(&v));
		ASSERT(v.vt == VT_I4);
		mediaType = static_cast<IMAPI_MEDIA_PHYSICAL_TYPE>(v.lVal);
		VariantClear(&v);
	}
	return mediaType;
}

ULONG CDiscFormatData::GetTotalSupportedWriteSpeeds()
{
	ASSERT(pDF2Data != NULL);
	CString lastErrorMessage;

	if(pSupportedWriteSpeeds) {
		SafeArrayDestroy(pSupportedWriteSpeeds);
		pSupportedWriteSpeeds = NULL;
	}

	HRESULT lastError = pDF2Data->get_SupportedWriteSpeeds(&pSupportedWriteSpeeds);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETSUPPORTEDWRITESPEEDS, lastError);
		SetLastError(lastError, lastErrorMessage);
		return 0;
	}
	SetLastError(lastError);
	return pSupportedWriteSpeeds->rgsabound[0].cElements;
}

ULONG CDiscFormatData::GetSupportedWriteSpeed(LONG index)
{
	if(!pSupportedWriteSpeeds)
		GetTotalSupportedWriteSpeeds();
	ASSERT(pSupportedWriteSpeeds != NULL);

	ULONG writeSpeed = 0;
	if(pSupportedWriteSpeeds) {
		VARIANT v;
		SafeArrayGetElement(pSupportedWriteSpeeds, &index, reinterpret_cast<LPVOID>(&v));
		ASSERT(v.vt == VT_I4);
		writeSpeed = v.lVal;
		VariantClear(&v);
	}
	return writeSpeed;
}

HRESULT CDiscFormatData::SetWriteSpeed(LONG writeSpeed)
{
	ASSERT(pDF2Data != NULL);
	HRESULT lastError = pDF2Data->SetWriteSpeed(writeSpeed, VARIANT_FALSE);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_SETWRITESPEED, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);
	return lastError;
}

LONG CDiscFormatData::GetFreeSectorsOnMedia()
{
	ASSERT(pDF2Data != NULL);
	CString lastErrorMessage;

	LONG sectors = -1;
	HRESULT lastError = pDF2Data->get_FreeSectorsOnMedia(&sectors);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETFREESECTORSONMEDIA, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);

	return sectors;
}

LONG CDiscFormatData::GetTotalSectorsOnMedia()
{
	ASSERT(pDF2Data != NULL);
	CString lastErrorMessage;

	LONG sectors = -1;
	HRESULT lastError = pDF2Data->get_TotalSectorsOnMedia(&sectors);
	if(FAILED(lastError)) {
		sectors = -1;
		lastErrorMessage.Format(IDS_ERR_GETTOTALSECTORSONMEDIA, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);

	return sectors;
}

HRESULT CDiscFormatData::EnableBurnProof(bool enable)
{
	ASSERT(pDF2Data != NULL);
	HRESULT lastError = pDF2Data->put_BufferUnderrunFreeDisabled(enable ? VARIANT_FALSE : VARIANT_TRUE);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_PUTBURNPROOF, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);
	return lastError;
}

HRESULT CDiscFormatData::SetCloseMedia(bool closeMedia)
{
	ASSERT(pDF2Data != NULL);
	HRESULT lastError = pDF2Data->put_ForceMediaToBeClosed(closeMedia ? VARIANT_TRUE : VARIANT_FALSE);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_PUTCLOSEMEDIA, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);
	return lastError;
}


HRESULT CDiscFormatData::GetMultisessionInterfaces(LPSAFEARRAY* ppMultiSession)
{
	ASSERT(pDF2Data != NULL);
	return pDF2Data->get_MultisessionInterfaces(ppMultiSession);
}