///////////////////////////////////////////////////////////////////////
// DiscRecorder.cpp
//
// Wrapper for IDiscRecorder2 interface

#include "stdafx.h"
#include "res/resource.h"
#include "DiscRecorder.h"


CDiscRecorder::CDiscRecorder(void) :
    pDiscRecorder(NULL),
    pVolumePathNames(NULL)
{
}

CDiscRecorder::~CDiscRecorder(void)
{
	if(pVolumePathNames)
		SafeArrayDestroy(pVolumePathNames);
	if(pDiscRecorder)
		pDiscRecorder->Release();
}


bool CDiscRecorder::Initialize(const CString& recorderUniqueID)
{
	ASSERT(pDiscRecorder == NULL);

	uniqueID = recorderUniqueID;

	CString lastErrorMessage;
	HRESULT lastError = CoCreateInstance(__uuidof(MsftDiscRecorder2), NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&pDiscRecorder));
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_INIT_DISCRECORDER2, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	CComBSTR bstr = uniqueID;
	lastError = pDiscRecorder->InitializeDiscRecorder(bstr);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_INITIALIZEDISCRECORDER, uniqueID, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	SetLastError(lastError);
	return true;
}

void CDiscRecorder::SetLastError(HRESULT lastError)
{
	this->lastError = lastError;
}

void CDiscRecorder::SetLastError(HRESULT lastError, CString lastErrorMessage)
{
	this->lastError = lastError;
	this->lastErrorMessage = lastErrorMessage;
}


LONG CDiscRecorder::GetLegacyDeviceNumber()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	LONG deviceNumber = 0;
	HRESULT lastError = pDiscRecorder->get_LegacyDeviceNumber(&deviceNumber);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETLEGACYDEVICENUMBER, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);
	return deviceNumber;
}

CString CDiscRecorder::GetProductID()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	BSTR productID = NULL;
	HRESULT lastError = pDiscRecorder->get_ProductId(&productID);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETPRODUCTID, lastError);
		SetLastError(lastError, lastErrorMessage);
		return TEXT("");
	} else
		SetLastError(lastError);

	CString s = productID;
	SysFreeString(productID);
	return s;
}

CString CDiscRecorder::GetProductRevision()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	BSTR productRevision = NULL;
	HRESULT lastError = pDiscRecorder->get_ProductRevision(&productRevision);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETPRODUCTREVISION, lastError);
		SetLastError(lastError, lastErrorMessage);
		return TEXT("");
	} else
		SetLastError(lastError);

	CString s = productRevision;
	SysFreeString(productRevision);
	return s;
}

CString CDiscRecorder::GetVendorID()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	BSTR vendorID = NULL;
	HRESULT lastError = pDiscRecorder->get_VendorId(&vendorID);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETVENDORID, lastError);
		SetLastError(lastError, lastErrorMessage);
		return TEXT("");
	} else
		SetLastError(lastError);

	CString s = vendorID;
	SysFreeString(vendorID);
	return s;
}

CString CDiscRecorder::GetVolumeName()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	BSTR volumeName = NULL;
	HRESULT lastError = pDiscRecorder->get_VolumeName(&volumeName);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETVOLUMENAME_RECORDER, lastError);
		SetLastError(lastError, lastErrorMessage);
		return TEXT("");
	} else
		SetLastError(lastError);

	CString s = volumeName;
	SysFreeString(volumeName);
	return s;
}

ULONG CDiscRecorder::GetTotalVolumePaths()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	if(pVolumePathNames) {
		SafeArrayDestroy(pVolumePathNames);
		pVolumePathNames = NULL;
	}

	HRESULT lastError = pDiscRecorder->get_VolumePathNames(&pVolumePathNames);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETVOLUMEPATHNAMES, lastError);
		SetLastError(lastError, lastErrorMessage);
		return 0;
	}
	SetLastError(lastError);
	return pVolumePathNames->rgsabound[0].cElements;
}

CString CDiscRecorder::GetVolumePath(LONG index)
{
	if(!pVolumePathNames)
		GetTotalVolumePaths();
	ASSERT(pVolumePathNames != NULL);

	CString s;
	if(pVolumePathNames) {
		VARIANT v;
		SafeArrayGetElement(pVolumePathNames, &index, reinterpret_cast<LPVOID>(&v));
		ASSERT(v.vt == VT_BSTR);
		s = v.bstrVal;
		VariantClear(&v);
	}
	return s;
}


bool CDiscRecorder::AcquireExclusiveAccess(bool force, const CString& clientName)
{
	ASSERT(pDiscRecorder != NULL);
	ASSERT(!clientName.IsEmpty());
	ASSERT(clientName.GetLength() < CDROM_EXCLUSIVE_CALLER_LENGTH);
	for(int i = 0; i < clientName.GetLength(); i++) {
		TCHAR ch = clientName[i];
		if(_istalnum(ch) || ch == TEXT(' ') || ch == TEXT('.') || ch == TEXT(',') || ch == TEXT(':') || ch == TEXT(';') || ch == TEXT('-') || ch == TEXT('_'))
			continue;

		ASSERT(FALSE && "Client name does not meet specification");
		return false;
	}

	CString lastErrorMessage;

	CComBSTR bstr = clientName;
	HRESULT lastError = pDiscRecorder->AcquireExclusiveAccess(force ? VARIANT_TRUE : VARIANT_FALSE, bstr);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_ACQUIREEXCLUSIVEACCESS, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	} else
		SetLastError(lastError);
	return true;
}

bool CDiscRecorder::ReleaseExclusiveAccess()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	HRESULT lastError = pDiscRecorder->ReleaseExclusiveAccess();
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_RELEASEEXCLUSIVEACCESS, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	} else
		SetLastError(lastError);
	return true;
}

CString CDiscRecorder::GetExclusiveAccessOwner()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	BSTR owner = NULL;
	HRESULT lastError = pDiscRecorder->get_ExclusiveAccessOwner(&owner);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETEXCLUSIVEACCESSOWNER, lastError);
		SetLastError(lastError, lastErrorMessage);
		return TEXT("");
	} else
		SetLastError(lastError);

	CString s = owner;
	SysFreeString(owner);
	return s;
}

bool CDiscRecorder::EjectMedia()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	HRESULT lastError = pDiscRecorder->EjectMedia();
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_EJECTMEDIA, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	SetLastError(lastError);
	return true;
}

bool CDiscRecorder::CloseTray()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	HRESULT lastError = pDiscRecorder->CloseTray();
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_CLOSETRAY, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	SetLastError(lastError);
	return true;
}

bool CDiscRecorder::EnableMcn()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	HRESULT lastError = pDiscRecorder->EnableMcn();
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_ENABLEMCN, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	SetLastError(lastError);
	return true;
}

bool CDiscRecorder::DisableMcn()
{
	ASSERT(pDiscRecorder != NULL);
	CString lastErrorMessage;

	HRESULT lastError = pDiscRecorder->DisableMcn();
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_DISABLEMCN, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	SetLastError(lastError);
	return true;
}