#include "stdafx.h"
#include "res/resource.h"
#include "ErasingController.h"

#include "DiscFormatEraseEvent.h"


CErasingController::CErasingController(void) :
    pWorkerThread(NULL),
    lastError(S_OK),
    hNotificationWnd(NULL),
    fullErase(VARIANT_TRUE),
    ejectMedia(false),
    cancel(false)
{
}

CErasingController::~CErasingController(void)
{
	if(pWorkerThread)
		WaitForSingleObject(*pWorkerThread, INFINITE);
}


bool CErasingController::GetCancelFlag()
{
	CSingleLock singleLock(&criticalSection);
	return cancel;
}

void CErasingController::SetCancelFlag(bool cancel)
{
	CSingleLock singleLock(&criticalSection);
	this->cancel = cancel;
}

void CErasingController::SetLastError(HRESULT lastError)
{
	CSingleLock singleLock(&criticalSection);
	this->lastError = lastError;
}

void CErasingController::SetLastError(HRESULT lastError, CString lastErrorMessage)
{
	CSingleLock singleLock(&criticalSection);
	this->lastError = lastError;
	this->lastErrorMessage = lastErrorMessage;
}


bool CErasingController::Erase()
{
	ASSERT(recorderID != IMAGE_RECORDER_ID);

	PTHREAD_PARAMS_ERASE pThreadData = new THREAD_PARAMS_ERASE;
	pThreadData->pThis = this;
	pThreadData->recorderID = recorderID;
	pThreadData->clientName = clientName;
	pThreadData->hNotificationWnd = hNotificationWnd;
	pThreadData->fullErase = fullErase;
	pThreadData->ejectMedia = ejectMedia;

	SetCancelFlag(false);
	pWorkerThread = AfxBeginThread(EraseThread, pThreadData, THREAD_PRIORITY_NORMAL);
	if(!pWorkerThread) {
		delete pThreadData;

		CString s;
		s.LoadString(IDS_ERR_ERASETHREAD);
		SetLastError(E_FAIL, s);
		return false;
	}
	SetLastError(S_OK);
	return true;
}

UINT CErasingController::EraseThread(LPVOID pParam)
{
	PTHREAD_PARAMS_ERASE pThreadData = reinterpret_cast<PTHREAD_PARAMS_ERASE>(pParam);

	HRESULT lastError = E_FAIL;
	CString lastErrorMessage;
	CDiscFormatEraseEvent* pEventSink = NULL;
	CDiscRecorder recorder;
	CDiscFormatErase eraser;

	PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_BEGIN_ERASE, 0, 0);

	if(!recorder.Initialize(pThreadData->recorderID)) {
		lastErrorMessage.LoadString(IDS_ERR_INITIALIZE_RECORDER);
		pThreadData->pThis->SetLastError(E_FAIL, lastErrorMessage);
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_ERASE, lastError, 0);
		goto CleanUp;
	}
	if(!eraser.Initialize(&recorder, pThreadData->clientName, pThreadData->hNotificationWnd)) {
		pThreadData->pThis->SetLastError(eraser.GetHresult(), eraser.GetErrorMessage());
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_ERASE, eraser.GetHresult(), 0);
		goto CleanUp;
	}

	pEventSink = CDiscFormatEraseEvent::CreateEventSink();
	if(!pEventSink) {
		lastErrorMessage.LoadString(IDS_ERR_CREATEEVENTSINK_ERASE);
		pThreadData->pThis->SetLastError(lastError, lastErrorMessage);
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_ERASE, lastError, 0);
		goto CleanUp;
	}
	if(!pEventSink->ConnectDiscFormatErase(&eraser)) {
		lastErrorMessage.LoadString(IDS_ERR_ADVISEEVENTSINK_ERASE);
		pThreadData->pThis->SetLastError(lastError, lastErrorMessage);
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_ERASE, lastError, 0);
		goto CleanUp;
	}
	pEventSink->SetNotificationWindow(pThreadData->hNotificationWnd);

	lastError = eraser.SetEraseMethod(pThreadData->fullErase);
	if(FAILED(lastError)) {
		pThreadData->pThis->SetLastError(lastError, eraser.GetErrorMessage());
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_ERASE, lastError, 0);
		goto CleanUp;
	}

	if(pThreadData->pThis->GetCancelFlag()) {
		lastErrorMessage.LoadString(IDS_ERR_OPERATIONCANCELED);
		pThreadData->pThis->SetLastError(S_OK, lastErrorMessage);
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_ERASE, S_FALSE, 0);
		goto CleanUp;
	}

	recorder.AcquireExclusiveAccess(true, pThreadData->clientName);
	lastError = eraser.GetInterface()->EraseMedia();
	if(FAILED(lastError)) {
		if(lastError == E_IMAPI_REQUEST_CANCELLED) {
			lastErrorMessage.LoadString(IDS_ERR_ERASING_CANCELED);
			pThreadData->pThis->SetLastError(S_OK, lastErrorMessage);
			PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_ERASE, S_FALSE, 0);
		} else {
			lastErrorMessage.Format(IDS_ERR_ERASEMEDIA, lastError);
			pThreadData->pThis->SetLastError(lastError, lastErrorMessage);
			PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_ERASE, lastError, 0);
		}
		goto CleanUp;
	}

	pThreadData->pThis->SetLastError(S_OK);
	PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_ERASE, S_OK, 0);

CleanUp:
	recorder.ReleaseExclusiveAccess();
	if(pThreadData->ejectMedia)
		recorder.EjectMedia();
	if(pEventSink)
		delete pEventSink;
	pThreadData->pThis->pWorkerThread = NULL;
	delete pThreadData;
	return 0;
}