///////////////////////////////////////////////////////////////////////
// FileSystemImage.cpp
//
// Wrapper for IFileSystemImage Interface

#include "stdafx.h"
#include "res/resource.h"
#include "FileSystemImage.h"

#include "FileSystemImageEvent.h"


CFileSystemImage::CFileSystemImage(void) :
    pWorkerThread(NULL),
    pFSImage(NULL),
    lastError(S_OK),
    hNotificationWnd(NULL),
    pSupportedISOLevels(NULL),
    pSupportedUDFRevisions(NULL),
    cancel(false)
{
}

CFileSystemImage::~CFileSystemImage(void)
{
	if(pWorkerThread)
		WaitForSingleObject(*pWorkerThread, INFINITE);

	if(pSupportedISOLevels)
		SafeArrayDestroy(pSupportedISOLevels);
	if(pSupportedUDFRevisions)
		SafeArrayDestroy(pSupportedUDFRevisions);
	if(pFSImage)
		pFSImage->Release();
}


bool CFileSystemImage::Initialize(HWND hNotificationWnd)
{
	ASSERT(pFSImage == NULL);
	this->hNotificationWnd = hNotificationWnd;

	CString lastErrorMessage;
	HRESULT lastError = CoCreateInstance(__uuidof(MsftFileSystemImage), NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&pFSImage));
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_INIT_FILESYSTEMIMAGE, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}

	lastError = pFSImage->put_FreeMediaBlocks(0);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_PUTFREEMEDIABLOCKS, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	CString s = CTime::GetCurrentTime().Format(TEXT("%d%m%y_%H%M"));
	if(!SetVolumeName(s))
		return false;

	SetLastError(lastError);
	return true;
}

bool CFileSystemImage::GetCancelFlag()
{
	CSingleLock singleLock(&criticalSection);
	return cancel;
}

void CFileSystemImage::SetCancelFlag(bool cancel)
{
	CSingleLock singleLock(&criticalSection);
	this->cancel = cancel;
}

void CFileSystemImage::SetLastError(HRESULT lastError)
{
	CSingleLock singleLock(&criticalSection);
	this->lastError = lastError;
}

void CFileSystemImage::SetLastError(HRESULT lastError, CString lastErrorMessage)
{
	CSingleLock singleLock(&criticalSection);
	this->lastError = lastError;
	this->lastErrorMessage = lastErrorMessage;
}


CString CFileSystemImage::GetVolumeName()
{
	ASSERT(pFSImage != NULL);
	CString lastErrorMessage;

	BSTR volumeName = NULL;
	HRESULT lastError = pFSImage->get_VolumeName(&volumeName);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETVOLUMENAME_IMAGE, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);

	CString s = volumeName;
	SysFreeString(volumeName);
	return s;
}

bool CFileSystemImage::SetVolumeName(CString& volumeName)
{
	ASSERT(pFSImage != NULL);
	CString lastErrorMessage;

	CComBSTR bstr = volumeName;
	HRESULT lastError = pFSImage->put_VolumeName(bstr);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_PUTVOLUMENAME, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	SetLastError(lastError);
	return true;
}

FsiFileSystems CFileSystemImage::GetSupportedFileSystems()
{
	ASSERT(pFSImage != NULL);
	CString lastErrorMessage;

	FsiFileSystems fs = FsiFileSystemNone;
	HRESULT lastError = pFSImage->get_FileSystemsSupported(&fs);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETFILESYSTEMSSUPPORTED, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);
	return fs;
}

bool CFileSystemImage::SetFileSystemsToCreate(FsiFileSystems fileSystem)
{
	ASSERT(pFSImage != NULL);
	CString lastErrorMessage;

	HRESULT lastError = pFSImage->put_FileSystemsToCreate(fileSystem);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_PUTFILESYSTEMSTOCREATE, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	SetLastError(lastError);
	return true;
}

ULONG CFileSystemImage::GetTotalSupportedISOLevels()
{
	ASSERT(pFSImage != NULL);
	CString lastErrorMessage;

	if(pSupportedISOLevels) {
		SafeArrayDestroy(pSupportedISOLevels);
		pSupportedISOLevels = NULL;
	}

	HRESULT lastError = pFSImage->get_ISO9660InterchangeLevelsSupported(&pSupportedISOLevels);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETISOLEVELSSUPPORTED, lastError);
		SetLastError(lastError, lastErrorMessage);
		return 0;
	}
	SetLastError(lastError);
	return pSupportedISOLevels->rgsabound[0].cElements;
}

LONG CFileSystemImage::GetSupportedISOLevel(LONG index)
{
	if(!pSupportedISOLevels)
		GetTotalSupportedISOLevels();
	ASSERT(pSupportedISOLevels != NULL);

	ULONG level = 0;
	if(pSupportedISOLevels) {
		VARIANT v;
		SafeArrayGetElement(pSupportedISOLevels, &index, reinterpret_cast<LPVOID>(&v));
		ASSERT(v.vt == VT_I4);
		level = v.lVal;
		VariantClear(&v);
	}
	return level;
}

bool CFileSystemImage::SetISOLevel(LONG level)
{
	ASSERT(pFSImage != NULL);
	CString lastErrorMessage;

	HRESULT lastError = pFSImage->put_ISO9660InterchangeLevel(level);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_PUTISOLEVEL, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	SetLastError(lastError);
	return true;
}

ULONG CFileSystemImage::GetTotalSupportedUDFRevisions()
{
	ASSERT(pFSImage != NULL);
	CString lastErrorMessage;

	if(pSupportedUDFRevisions) {
		SafeArrayDestroy(pSupportedUDFRevisions);
		pSupportedUDFRevisions = NULL;
	}

	HRESULT lastError = pFSImage->get_UDFRevisionsSupported(&pSupportedUDFRevisions);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETUDFREVISIONSSUPPORTED, lastError);
		SetLastError(lastError, lastErrorMessage);
		return 0;
	}
	SetLastError(lastError);
	return pSupportedUDFRevisions->rgsabound[0].cElements;
}

LONG CFileSystemImage::GetSupportedUDFRevision(LONG index)
{
	if(!pSupportedUDFRevisions)
		GetTotalSupportedUDFRevisions();
	ASSERT(pSupportedISOLevels != NULL);

	LONG revision = 0;
	if(pSupportedUDFRevisions) {
		VARIANT v;
		SafeArrayGetElement(pSupportedUDFRevisions, &index, reinterpret_cast<LPVOID>(&v));
		ASSERT(v.vt == VT_I4);
		revision = v.lVal;
		VariantClear(&v);
	}
	return revision;
}

bool CFileSystemImage::SetUDFRevision(LONG revision)
{
	ASSERT(pFSImage != NULL);
	CString lastErrorMessage;

	HRESULT lastError = pFSImage->put_UDFRevision(revision);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_PUTUDFREVISION, lastError);
		SetLastError(lastError, lastErrorMessage);
		return false;
	}
	SetLastError(lastError);
	return true;
}

LONG CFileSystemImage::GetBlocksInUse()
{
	ASSERT(pFSImage != NULL);
	CString lastErrorMessage;

	LONG blocksInUse = -1;
	HRESULT lastError = pFSImage->get_UsedBlocks(&blocksInUse);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETUSEDBLOCKS, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);
	return blocksInUse;
}

HRESULT CFileSystemImage::GetRoot(IFsiDirectoryItem** ppRootDirectory)
{
	ASSERT(pFSImage != NULL);
	HRESULT lastError = pFSImage->get_Root(ppRootDirectory);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETROOT, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);
	return lastError;
}

HRESULT CFileSystemImage::CreateDirectoryItem(CString itemName, IFsiDirectoryItem** ppNewItem)
{
	ASSERT(pFSImage != NULL);
	CComBSTR bstr = itemName;
	HRESULT lastError = pFSImage->CreateDirectoryItem(bstr, ppNewItem);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_CREATEDIRECTORYITEM, itemName, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);
	return lastError;
}

LPSTREAM CFileSystemImage::GetStream(IFileSystemImageResult** ppImageResult)
{
	ASSERT(pFSImage != NULL);
	CComPtr<IFileSystemImageResult> pImageResult = NULL;
	HRESULT lastError = pFSImage->CreateResultImage(&pImageResult);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_CREATERESULTSTREAM, lastError);
		SetLastError(lastError, lastErrorMessage);
		return NULL;
	}
	LPSTREAM pData = NULL;
	lastError = pImageResult->get_ImageStream(&pData);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETIMAGESTREAM, lastError);
		SetLastError(lastError, lastErrorMessage);
		return NULL;
	}

	if(ppImageResult)
		pImageResult->QueryInterface(IID_PPV_ARGS(ppImageResult));
	SetLastError(lastError);
	return pData;
}

LONG CFileSystemImage::GetTotalDirectories()
{
	ASSERT(pFSImage != NULL);

	LONG c = 0;
	HRESULT lastError = pFSImage->get_DirectoryCount(&c);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETDIRECTORYCOUNT, lastError);
		SetLastError(lastError, lastErrorMessage);
		return -1;
	}
	SetLastError(lastError);
	return c;
}

LONG CFileSystemImage::GetTotalFiles()
{
	ASSERT(pFSImage != NULL);

	LONG c = 0;
	HRESULT lastError = pFSImage->get_FileCount(&c);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETFILECOUNT, lastError);
		SetLastError(lastError, lastErrorMessage);
		return -1;
	}
	SetLastError(lastError);
	return c;
}


HRESULT CFileSystemImage::SetMultisessionInterfaces(LPSAFEARRAY pMultiSession)
{
	ASSERT(pFSImage != NULL);
	return pFSImage->put_MultisessionInterfaces(pMultiSession);
}

HRESULT CFileSystemImage::ImportDefaultSession()
{
	ASSERT(pFSImage != NULL);
	FsiFileSystems fs;
	return pFSImage->ImportFileSystem(&fs);
}

/*CString CFileSystemImage::GetImportedVolumeName()
{
	ASSERT(pFSImage != NULL);
	CString lastErrorMessage;

	BSTR volumeName = NULL;
	HRESULT lastError = pFSImage->get_ImportedVolumeName(&volumeName);
	if(FAILED(lastError)) {
		lastErrorMessage.Format(IDS_ERR_GETIMPORTEDVOLUMENAME_IMAGE, lastError);
		SetLastError(lastError, lastErrorMessage);
	} else
		SetLastError(lastError);

	CString s = volumeName;
	SysFreeString(volumeName);
	return s;
}*/


bool CFileSystemImage::AddFile(IFsiDirectoryItem* pBaseDirectory, CString& file)
{
	CArray<CString> files;
	files.Add(file);
	return AddFiles(pBaseDirectory, files, PathIsDirectory(file));
}

bool CFileSystemImage::AddFiles(IFsiDirectoryItem* pBaseDirectory, CArray<CString>& files, BOOL mixedData)
{
	ASSERT(pFSImage != NULL);

	PTHREAD_PARAMS_ADDFILES pData = new THREAD_PARAMS_ADDFILES;
	pData->pThis = this;
	pData->files.Append(files);
	pData->mixedData = mixedData;
	pBaseDirectory->QueryInterface(IID_PPV_ARGS(&pData->pBaseDirectory));

	SetCancelFlag(false);
	pWorkerThread = AfxBeginThread(AddFilesThread, pData, THREAD_PRIORITY_NORMAL);
	if(!pWorkerThread) {
		pData->pBaseDirectory->Release();
		delete pData;

		CString s;
		s.LoadString(IDS_ERR_ADDFILESTHREAD);
		SetLastError(E_FAIL, s);
		return false;
	}
	SetLastError(S_OK);
	return true;
}

bool CFileSystemImage::AddTree(IFsiDirectoryItem* pBaseDirectory, CString& dirPath, VARIANT_BOOL includeBaseDirectory)
{
	ASSERT(pFSImage != NULL);

	PTHREAD_PARAMS_ADDTREE pData = new THREAD_PARAMS_ADDTREE;
	pData->pThis = this;
	pData->dirPath = dirPath;
	pBaseDirectory->QueryInterface(IID_PPV_ARGS(&pData->pBaseDirectory));
	pData->includeBaseDirectory = includeBaseDirectory;

	SetCancelFlag(false);
	pWorkerThread = AfxBeginThread(AddTreeThread, pData, THREAD_PRIORITY_NORMAL);
	if(!pWorkerThread) {
		pData->pBaseDirectory->Release();
		delete pData;

		CString s;
		s.LoadString(IDS_ERR_ADDTREETHREAD);
		SetLastError(E_FAIL, s);
		return false;
	}
	SetLastError(S_OK);
	return true;
}

UINT CFileSystemImage::AddFilesThread(LPVOID pParam)
{
	PTHREAD_PARAMS_ADDFILES pData = reinterpret_cast<PTHREAD_PARAMS_ADDFILES>(pParam);

	HRESULT lastError = E_FAIL;
	CString lastErrorMessage;
	CFileSystemImageEvent* pEventSink = NULL;
	UINT dataType = DATATYPE_FILE;
	if(pData->mixedData)
		dataType |= DATATYPE_TREE;

	SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_BEGIN_ADD_DATA, 0, dataType);
	if(pData->pThis->GetCancelFlag()) {
		lastErrorMessage.LoadString(IDS_ERR_OPERATIONCANCELED);
		pData->pThis->SetLastError(S_OK, lastErrorMessage);
		SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, S_FALSE, dataType);
		goto CleanUp;
	}

	pEventSink = CFileSystemImageEvent::CreateEventSink();
	if(!pEventSink) {
		lastErrorMessage.LoadString(IDS_ERR_CREATEEVENTSINK_IMAGE);
		pData->pThis->SetLastError(lastError, lastErrorMessage);
		SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, lastError, dataType);
		goto CleanUp;
	}
	if(!pEventSink->ConnectFileSystemImage(pData->pThis)) {
		lastErrorMessage.LoadString(IDS_ERR_ADVISEEVENTSINK_IMAGE);
		pData->pThis->SetLastError(lastError, lastErrorMessage);
		SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, lastError, dataType);
		goto CleanUp;
	}
	pEventSink->SetNotificationWindow(pData->pThis->hNotificationWnd);

	for(INT_PTR i = 0; i < pData->files.GetCount(); i++) {
		BOOL isDirectory = PathIsDirectory(pData->files[i]);

		if(pData->pThis->GetCancelFlag()) {
			lastErrorMessage.LoadString(IDS_ERR_OPERATIONCANCELED);
			pData->pThis->SetLastError(S_OK, lastErrorMessage);
			SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, S_FALSE, dataType);
			goto CleanUp;
		}

		if(isDirectory) {
			pEventSink->SetCurrentFile(TEXT(""));
			CComBSTR bstr = pData->files[i];
			if(PathIsRoot(pData->files[i])) {
				CComPtr<IFsiDirectoryItem> pDriveDirectory;
				lastError = pData->pThis->CreateDirectoryItem(pData->files[i].Left(1), &pDriveDirectory);
				if(FAILED(lastError)) {
					pData->pThis->SetLastError(lastError, pData->pThis->GetErrorMessage());
					SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, lastError, dataType);
					goto CleanUp;
				}
				CTime t = CTime::GetCurrentTime();
				SYSTEMTIME st;
				t.GetAsSystemTime(st);
				COleDateTime creationTime(st);
				pDriveDirectory->put_CreationTime(creationTime.m_dt);
				pDriveDirectory->put_LastModifiedTime(creationTime.m_dt);
				pDriveDirectory->put_LastAccessedTime(creationTime.m_dt);

				lastError = pData->pBaseDirectory->Add(pDriveDirectory);
				if(FAILED(lastError)) {
					lastErrorMessage.Format(IDS_ERR_ADD, pData->files[i], lastError);
					pData->pThis->SetLastError(lastError, lastErrorMessage);
					SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, lastError, dataType);
					goto CleanUp;
				}
				lastError = pDriveDirectory->AddTree(bstr, VARIANT_FALSE);
			} else
				lastError = pData->pBaseDirectory->AddTree(bstr, VARIANT_TRUE);
			if(FAILED(lastError)) {
				lastErrorMessage.Format(IDS_ERR_ADDTREE, pData->files[i], lastError);
				pData->pThis->SetLastError(lastError, lastErrorMessage);
				SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, lastError, dataType);
				goto CleanUp;
			}
		} else {
			pEventSink->SetCurrentFile(pData->files[i]);
			CComPtr<IStream> pStream;
			lastError = SHCreateStreamOnFileEx(pData->files[i], STGM_READ | STGM_SHARE_DENY_WRITE, FILE_ATTRIBUTE_NORMAL, FALSE, NULL, &pStream);
			if(FAILED(lastError)) {
				lastErrorMessage.Format(IDS_ERR_COULD_NOT_READ_FILE, pData->files[i], lastError);
				pData->pThis->SetLastError(lastError, lastErrorMessage);
				SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, lastError, dataType);
				goto CleanUp;
			}
			if(pStream) {
				CComBSTR bstr = PathFindFileName(pData->files[i]);
				lastError = pData->pBaseDirectory->AddFile(bstr, pStream);
				if(FAILED(lastError)) {
					lastErrorMessage.Format(IDS_ERR_ADDFILE, pData->files[i], lastError);
					pData->pThis->SetLastError(lastError, lastErrorMessage);
					SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, lastError, dataType);
					goto CleanUp;
				}

				// retrieve last modified time
				HANDLE hFile = CreateFile(pData->files[i], GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
				if(hFile != INVALID_HANDLE_VALUE) {
					FILETIME ft;
					GetFileTime(hFile, NULL, NULL, &ft);
					CloseHandle(hFile);
					COleDateTime creationTime(ft);
					CComPtr<IFsiItem> pItem;
					pData->pBaseDirectory->get_Item(bstr, &pItem);
					if(pItem) {
						pItem->put_CreationTime(creationTime.m_dt);
						pItem->put_LastModifiedTime(creationTime.m_dt);
						pItem->put_LastAccessedTime(creationTime.m_dt);
					}
				}
			}
		}
	}

	pData->pThis->SetLastError(S_OK);
	SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, S_OK, dataType);

CleanUp:
	pData->pBaseDirectory->Release();
	if(pEventSink)
		delete pEventSink;
	pData->pThis->pWorkerThread = NULL;
	delete pData;
	return 0;
}

UINT CFileSystemImage::AddTreeThread(LPVOID pParam)
{
	PTHREAD_PARAMS_ADDTREE pData = reinterpret_cast<PTHREAD_PARAMS_ADDTREE>(pParam);

	HRESULT lastError = E_FAIL;
	CString lastErrorMessage;
	CFileSystemImageEvent* pEventSink = NULL;

	SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_BEGIN_ADD_DATA, 0, DATATYPE_TREE);
	if(pData->pThis->GetCancelFlag()) {
		lastErrorMessage.LoadString(IDS_ERR_OPERATIONCANCELED);
		pData->pThis->SetLastError(S_OK, lastErrorMessage);
		SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, S_FALSE, DATATYPE_TREE);
		goto CleanUp;
	}

	pEventSink = CFileSystemImageEvent::CreateEventSink();
	if(!pEventSink) {
		lastErrorMessage.LoadString(IDS_ERR_CREATEEVENTSINK_IMAGE);
		pData->pThis->SetLastError(lastError, lastErrorMessage);
		SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, lastError, DATATYPE_TREE);
		goto CleanUp;
	}
	if(!pEventSink->ConnectFileSystemImage(pData->pThis)) {
		lastErrorMessage.LoadString(IDS_ERR_ADVISEEVENTSINK_IMAGE);
		pData->pThis->SetLastError(lastError, lastErrorMessage);
		SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, lastError, DATATYPE_TREE);
		goto CleanUp;
	}
	pEventSink->SetNotificationWindow(pData->pThis->hNotificationWnd);

	if(pData->pThis->GetCancelFlag()) {
		lastErrorMessage.LoadString(IDS_ERR_OPERATIONCANCELED);
		pData->pThis->SetLastError(S_OK, lastErrorMessage);
		SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, S_FALSE, DATATYPE_TREE);
		goto CleanUp;
	}
	// for some reason, CComBSTR doesn't like the "CleanUp" label
	{
		CComBSTR bstr = pData->dirPath;
		lastError = pData->pBaseDirectory->AddTree(bstr, pData->includeBaseDirectory);
		if(FAILED(lastError)) {
			lastErrorMessage.Format(IDS_ERR_ADDTREE, pData->dirPath, lastError);
			pData->pThis->SetLastError(lastError, lastErrorMessage);
			SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, lastError, DATATYPE_TREE);
			goto CleanUp;
		}
	}

	pData->pThis->SetLastError(S_OK);
	SendMessage(pData->pThis->hNotificationWnd, WM_IMAPI_END_ADD_DATA, S_OK, DATATYPE_TREE);

CleanUp:
	pData->pBaseDirectory->Release();
	if(pEventSink)
		delete pEventSink;
	pData->pThis->pWorkerThread = NULL;
	delete pData;
	return 0;
}