#include "stdafx.h"
#include "res/resource.h"
#include "RecordingController.h"

#include "DiscFormatDataEvent.h"


CRecordingController::CRecordingController(void) :
    pWorkerThread(NULL),
    lastError(S_OK),
    hNotificationWnd(NULL),
    enableBurnProof(true),
    closeMedia(true),
    ejectMedia(false),
    writeSpeed(-1),
    cancel(false)
{
}

CRecordingController::~CRecordingController(void)
{
	if(pWorkerThread)
		WaitForSingleObject(*pWorkerThread, INFINITE);
}


bool CRecordingController::GetCancelFlag()
{
	CSingleLock singleLock(&criticalSection);
	return cancel;
}

void CRecordingController::SetCancelFlag(bool cancel)
{
	CSingleLock singleLock(&criticalSection);
	this->cancel = cancel;
}

void CRecordingController::SetLastError(HRESULT lastError)
{
	CSingleLock singleLock(&criticalSection);
	this->lastError = lastError;
}

void CRecordingController::SetLastError(HRESULT lastError, CString lastErrorMessage)
{
	CSingleLock singleLock(&criticalSection);
	this->lastError = lastError;
	this->lastErrorMessage = lastErrorMessage;
}


bool CRecordingController::Burn(LPSTREAM pData)
{
	PTHREAD_PARAMS_BURN pThreadData = new THREAD_PARAMS_BURN;
	pThreadData->pThis = this;
	pThreadData->recorderID = recorderID;
	pThreadData->clientName = clientName;
	pThreadData->hNotificationWnd = hNotificationWnd;
	pThreadData->enableBurnProof = enableBurnProof;
	pThreadData->closeMedia = closeMedia;
	pThreadData->ejectMedia = ejectMedia;
	pThreadData->writeSpeed = writeSpeed;
	pData->QueryInterface(IID_PPV_ARGS(&pThreadData->pData));

	SetCancelFlag(false);
	if(recorderID == IMAGE_RECORDER_ID) {
		pThreadData->volumeName = volumeName;
		if(pThreadData->volumeName.IsEmpty())
			pThreadData->volumeName = TEXT("Image.iso");
		pThreadData->totalBlocks = totalBlocks;
		pThreadData->blockSize = blockSize;
		pWorkerThread = AfxBeginThread(BurnThread_ImageRecorder, pThreadData, THREAD_PRIORITY_NORMAL);
	} else
		pWorkerThread = AfxBeginThread(BurnThread, pThreadData, THREAD_PRIORITY_NORMAL);
	if(!pWorkerThread) {
		pThreadData->pData->Release();
		delete pThreadData;

		CString s;
		s.LoadString(IDS_ERR_BURNTHREAD);
		SetLastError(E_FAIL, s);
		return false;
	}
	SetLastError(S_OK);
	return true;
}

UINT CRecordingController::BurnThread(LPVOID pParam)
{
	PTHREAD_PARAMS_BURN pThreadData = reinterpret_cast<PTHREAD_PARAMS_BURN>(pParam);

	HRESULT lastError = E_FAIL;
	CString lastErrorMessage;
	CDiscFormatDataEvent* pEventSink = NULL;
	CDiscRecorder recorder;
	CDiscFormatData dataWriter;

	PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_BEGIN_BURN, 0, 0);

	if(!recorder.Initialize(pThreadData->recorderID)) {
		lastErrorMessage.LoadString(IDS_ERR_INITIALIZE_RECORDER);
		pThreadData->pThis->SetLastError(E_FAIL, lastErrorMessage);
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, lastError, 0);
		goto CleanUp;
	}
	if(!dataWriter.Initialize(&recorder, pThreadData->clientName, pThreadData->hNotificationWnd)) {
		pThreadData->pThis->SetLastError(dataWriter.GetHresult(), dataWriter.GetErrorMessage());
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, dataWriter.GetHresult(), 0);
		goto CleanUp;
	}

	pEventSink = CDiscFormatDataEvent::CreateEventSink();
	if(!pEventSink) {
		lastErrorMessage.LoadString(IDS_ERR_CREATEEVENTSINK_DATA);
		pThreadData->pThis->SetLastError(lastError, lastErrorMessage);
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, lastError, 0);
		goto CleanUp;
	}
	if(!pEventSink->ConnectDiscFormatData(&dataWriter)) {
		lastErrorMessage.LoadString(IDS_ERR_ADVISEEVENTSINK_DATA);
		pThreadData->pThis->SetLastError(lastError, lastErrorMessage);
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, lastError, 0);
		goto CleanUp;
	}
	pEventSink->SetNotificationWindow(pThreadData->hNotificationWnd);

	lastError = dataWriter.EnableBurnProof(pThreadData->enableBurnProof);
	if(FAILED(lastError)) {
		pThreadData->pThis->SetLastError(lastError, dataWriter.GetErrorMessage());
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, lastError, 0);
		goto CleanUp;
	}
	lastError = dataWriter.SetCloseMedia(pThreadData->closeMedia);
	if(FAILED(lastError)) {
		pThreadData->pThis->SetLastError(lastError, dataWriter.GetErrorMessage());
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, lastError, 0);
		goto CleanUp;
	}
	lastError = dataWriter.SetWriteSpeed(pThreadData->writeSpeed);
	if(FAILED(lastError)) {
		pThreadData->pThis->SetLastError(lastError, dataWriter.GetErrorMessage());
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, lastError, 0);
		goto CleanUp;
	}

	if(pThreadData->pThis->GetCancelFlag()) {
		lastErrorMessage.LoadString(IDS_ERR_OPERATIONCANCELED);
		pThreadData->pThis->SetLastError(S_OK, lastErrorMessage);
		PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, S_FALSE, 0);
		goto CleanUp;
	}

	recorder.AcquireExclusiveAccess(true, pThreadData->clientName);
	lastError = dataWriter.GetInterface()->Write(pThreadData->pData);
	if(FAILED(lastError)) {
		if(lastError == E_IMAPI_REQUEST_CANCELLED) {
			lastErrorMessage.LoadString(IDS_ERR_RECORDING_CANCELED);
			pThreadData->pThis->SetLastError(S_OK, lastErrorMessage);
			PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, S_FALSE, 0);
		} else {
			lastErrorMessage.Format(IDS_ERR_WRITE, lastError);
			pThreadData->pThis->SetLastError(lastError, lastErrorMessage);
			PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, lastError, 0);
		}
		goto CleanUp;
	}

	pThreadData->pThis->SetLastError(S_OK);
	PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, S_OK, 0);

CleanUp:
	recorder.ReleaseExclusiveAccess();
	if(pThreadData->ejectMedia)
		recorder.EjectMedia();
	pThreadData->pData->Release();
	if(pEventSink)
		delete pEventSink;
	pThreadData->pThis->pWorkerThread = NULL;
	delete pThreadData;
	return 0;
}

UINT CRecordingController::BurnThread_ImageRecorder(LPVOID pParam)
{
	PTHREAD_PARAMS_BURN pThreadData = reinterpret_cast<PTHREAD_PARAMS_BURN>(pParam);
	CString defaultFileName = pThreadData->volumeName;
	if(!defaultFileName.IsEmpty())
		defaultFileName += TEXT(".iso");

	HRESULT lastError = E_FAIL;
	CString lastErrorMessage;

	PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_BEGIN_BURN, 0, 0);

	// let the user choose a file to write to
	CWnd parentWindow;
	parentWindow.Attach(pThreadData->hNotificationWnd);
	CString filters;
	filters.LoadString(IDS_OPEN_IMAGE_FILTERS);
	CFileDialog fileDlg(FALSE, TEXT("iso"), defaultFileName, OFN_DONTADDTORECENT | OFN_NOREADONLYRETURN | OFN_OVERWRITEPROMPT | OFN_PATHMUSTEXIST, filters, &parentWindow);
	if(IsVistaOrNewer())
		fileDlg.UpdateOFNFromShellDialog();

	if(fileDlg.DoModal() == IDOK) {
		CString fileName = fileDlg.GetPathName();
		lastError = WriteImage(fileName, pThreadData);
	} else
		lastError = E_IMAPI_REQUEST_CANCELLED;
	parentWindow.Detach();

	if(FAILED(lastError)) {
		if(lastError == E_IMAPI_REQUEST_CANCELLED) {
			lastErrorMessage.LoadString(IDS_ERR_RECORDING_CANCELED);
			pThreadData->pThis->SetLastError(S_OK, lastErrorMessage);
			PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, S_FALSE, 0);
		} else {
			lastErrorMessage.Format(IDS_ERR_WRITEIMAGE, lastError);
			pThreadData->pThis->SetLastError(lastError, lastErrorMessage);
			PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, lastError, 0);
		}
		goto CleanUp;
	}

	pThreadData->pThis->SetLastError(S_OK);
	PostMessage(pThreadData->hNotificationWnd, WM_IMAPI_END_BURN, S_OK, 0);

CleanUp:
	pThreadData->pData->Release();
	pThreadData->pThis->pWorkerThread = NULL;
	delete pThreadData;
	return 0;
}

HRESULT CRecordingController::WriteImage(const CString& fileName, PTHREAD_PARAMS_BURN pThreadData)
{
	LPSTREAM pTargetStream = NULL;
	HRESULT lastError = SHCreateStreamOnFileEx(fileName, STGM_CREATE | STGM_WRITE | STGM_SHARE_DENY_WRITE, FILE_ATTRIBUTE_NORMAL, FALSE, NULL, &pTargetStream);
	if(FAILED(lastError) || !pTargetStream) {
		if(pTargetStream)
			pTargetStream->Release();
		return lastError;
	}

	// write 100 blocks at once
	const int bufferSize = 100 * pThreadData->blockSize;
	LPVOID pBuffer = HeapAlloc(GetProcessHeap(), 0, bufferSize);
	if(!pBuffer) {
		if(pTargetStream)
			pTargetStream->Release();
		return E_OUTOFMEMORY;
	}

	IMAPI_FORMAT_DATA_UPDATE_EVENT_ARGS eventArgs;
	ZeroMemory(&eventArgs, sizeof(eventArgs));
	CTime startTime = CTime::GetCurrentTime();
	eventArgs.elapsedTime = 0;
	eventArgs.remainingTime = -1;
	eventArgs.sectorCount = pThreadData->totalBlocks;
	eventArgs.currentAction = IMAPI_FORMAT2_DATA_WRITE_ACTION_WRITING_DATA;
	ULONGLONG totalBytesWritten = 0;
	ULONG bytesTransfered;
	do {
		bytesTransfered = 0;
		lastError = pThreadData->pData->Read(pBuffer, bufferSize, &bytesTransfered);
		if(FAILED(lastError)) {
			eventArgs.currentAction = IMAPI_FORMAT2_DATA_WRITE_ACTION_FINALIZATION;
			SendMessage(pThreadData->hNotificationWnd, WM_IMAPI_FORMAT_DATA_UPDATE, 0, reinterpret_cast<LPARAM>(&eventArgs));
			pTargetStream->Commit(STGC_DEFAULT);
			pTargetStream->Release();
			HeapFree(GetProcessHeap(), 0, pBuffer);
			return lastError;
		}

		if(bytesTransfered > 0) {
			lastError = pTargetStream->Write(pBuffer, bufferSize, &bytesTransfered);
			if(FAILED(lastError)) {
				eventArgs.currentAction = IMAPI_FORMAT2_DATA_WRITE_ACTION_FINALIZATION;
				SendMessage(pThreadData->hNotificationWnd, WM_IMAPI_FORMAT_DATA_UPDATE, 0, reinterpret_cast<LPARAM>(&eventArgs));
				pTargetStream->Commit(STGC_DEFAULT);
				pTargetStream->Release();
				HeapFree(GetProcessHeap(), 0, pBuffer);
				return lastError;
			}
			totalBytesWritten += bytesTransfered;
			eventArgs.lastWrittenLba = static_cast<LONG>(totalBytesWritten / IMAPI_SECTOR_SIZE);

			CTimeSpan ts = CTime::GetCurrentTime() - startTime;
			eventArgs.elapsedTime = static_cast<LONG>(ts.GetTotalSeconds());
			LRESULT ret = SendMessage(pThreadData->hNotificationWnd, WM_IMAPI_FORMAT_DATA_UPDATE, 0, reinterpret_cast<LPARAM>(&eventArgs));
			if(ret == RETURN_CANCEL_WRITE) {
				eventArgs.currentAction = IMAPI_FORMAT2_DATA_WRITE_ACTION_FINALIZATION;
				SendMessage(pThreadData->hNotificationWnd, WM_IMAPI_FORMAT_DATA_UPDATE, 0, reinterpret_cast<LPARAM>(&eventArgs));
				pTargetStream->Commit(STGC_DEFAULT);
				pTargetStream->Release();
				HeapFree(GetProcessHeap(), 0, pBuffer);
				DeleteFile(fileName);
				return E_IMAPI_REQUEST_CANCELLED;
			}
		}
	} while(bytesTransfered > 0);

	eventArgs.currentAction = IMAPI_FORMAT2_DATA_WRITE_ACTION_FINALIZATION;
	SendMessage(pThreadData->hNotificationWnd, WM_IMAPI_FORMAT_DATA_UPDATE, 0, reinterpret_cast<LPARAM>(&eventArgs));
	pTargetStream->Commit(STGC_DEFAULT);
	pTargetStream->Release();
	HeapFree(GetProcessHeap(), 0, pBuffer);
	return S_OK;
}